
# Routing API design

## Scenario

For designing the API this fairly common and complex scenario will be taken as the base case to solve:

```
  '/loading' => user access this page when the authentication state is unknown
  '/sign-in' => user access this page when unauthenticated
  '/sign-in/phone' => user can access this nested page when unauthenticated 
  '/create-profile' => when user is authenticated but doesn't have a profile 
  '/dashboard' => when the user is authenticated & has a profile, the dashboard
  '/products' => when the user is authenticated & has a profile and clicked on a menu item to see the products
  '/products/:id' => when the user wants to see a specific product
```

## Route naming convention 

From the scenario we can notice that the up stack can be defined as a function of the url path where each segment is a screen in the stack.
For example when on the '/sign-in/phone' route the up stack will look like this:

```
  - SignInScreen
  - SignInWithPhoneScreen
```

On the product details page the up stack will look like this

```
  - ProductsScreen
  - ProductDetailsScreen
```

The only question mark is what happens with the path '/' that we will call home route.


## Home route

Websites are firstly accessed by their main url, eg: www.facebook.com. That main url we define it as `home route` or '/' for short.

The problem with the home route is that the material specs aren't realistic as a webapp standpoint in our scenario. The material specs specify that the home must be the upper most activity in the up stack. That doesn't play well with /sign-in page.

2 question marks remain for that home route:

  - Should the home route be part of the stack when navigating to '/products'
  - What happens when the user types the home route in our scenario since we didn't define that route


The home route issue will be kept on the side until an API design is picked. 

## API design ideas
### Option design 1 : Strict routing naming conventions

```
Router(
  routes: {
    '/sign-in': (context, params) => SignInPage(),
    '/sign-in/phone': (context, params) => SignInWithPhonePage(),
    '/create-profile': (context, params) => CreateProfilePage(),
    '/dashboard': (context, params) => Dashboard(),
    '/products': (context, params) => ProductsPage(),
    '/products/:id': (context, params) => ProductDetailsPage(params['id']),
  },

)
```

In this scenario the routing mechanism obey strict conventions of the url to manage the stack. As discussed before.

```
'/sign-in/phone': has SignInScreen and SignInWithPhoneScreen in the stack
'/dashboard': has dashboard in the stack
'/products/:id': has ProductDetailScreen and ProductsScreen in the stack
```

If the user wants the dashboard to be the top most Screen in the up stack of products, the user can change the routing as such:

```
Router(
  routes: {
    '/sign-in': (context, params) => SignInPage(),
    '/sign-in/phone': (context, params) => SignInWithPhonePage(),
    '/create-profile': (context, params) => CreateProfilePage(),
    '/app': (context, params) => DashboardPage(),
    '/app/products': (context, params) => ProductsPage(),
    '/app/products/:id': (context, params) => ProductDetailsPage(params['id']),
  }
)
```

*Drawbacks*: 
  - url might not be exactly to user preferences 

### Option design 2 : Nesting

```
Router(
  routes: [
    RouteConfig(
      path: '/sign-in',
      builder: (context, params) => SignInPage(),
      children: [
        RouteConfig( 
          path: '/phone', 
          builder: (context, params) => SignInWithPhonePage()
        )
      ]
    ),
    //...
  }
)
```

We can see that if the route naming convention has to be respected here too, it doesn't offer any advantage in readability. We have to check the parent to see the whole path

*Drawbacks*:
  - Less readability
  - harder to implement as we have to check parents (tree) instead of just checking the segments of a path.


### Option design 3: pattern matching


```
Router(
  routes: {
    '/sign-in': (context, params) => SignInPage(),
    '/sign-in/**': (context, params) => SignInPage(),
    '/sign-in/phone': (context, params) => SignInWithPhonePage(),
    '/create-profile': (context, params) => CreateProfilePage(),
    '/dashboard': (context, params) => DashBoardPage(),
    '/products': (context, params) => ProductsPage(),
    '/products/**': (context, params) => ProductsPage(),
    '/products/:id': (context, params) => ProductDetailsPage(params['id']),
  }
)
```

Here we have repetition but we could make a rule that ** allow the current directory as well or add another wildcard

```
Router(
  routes: {
    '/sign-in/?': (context, params) => SignInPage(),
    '/sign-in/phone': (context, params) => SignInWithPhonePage(),
    '/create-profile': (context, params) => CreateProfilePage(),
    '/dashboard': (context, params) => DashboardPage(),
    '/products/?': (context, params) => ProductsPage(),
    '/products/:id': (context, params) => ProductDetailsPage(params['id']),
  }
)
```


*advantages* :
  - the user has more control over the stack
*drawbacks* :
  - user has to learn a pattern


### Option design 4 Add stack properties to the routes 

```
Router(
  routes: [
    Route(
      path: '/sign-in', 
      builder: (context, params) => SignInPage(),
    ),
    Route(
      path: '/sign-in/phone', 
      builder: (context, params) => SignInWithPhonePage(),
      upstack: ['/sign-in'] 
    ),
  ]
)
```

 - Although a bit less consise, having a route class would make typing easier for future version.
 - It remains to be defined if custom stacks in this scenario would be useful.

A possible scenario would be: 
We could have /forms/:formId/submit-status and want an up buttons that lets the user go back to the list of forms but not modify the form he already submitted. The stack would look like ['forms'] (+ current route).

### Option 5: allow exact matches

```
Router(
  routes: [
    Route(
      path: '/sign-in', 
      builder: (context, params) => SignInPage(),
      exactMatch: true
    ),
    Route(
      path: '/sign-in/phone', 
      builder: (context, params) => SignInWithPhonePage(),
    ),
  ]
)
```

Here the user can set a route as exactMatch, and the SignInPage won't be in the upstack when the user is on page sign-in/phone.

It remains to be seen if this would be of any usefulness...


### Conclusions of options

While scenario 1 is really easy to setup, scenario 3 offers a bit more control as well as scenario 4 & 5. Scenario 4 introduce Route class that would make it more future proof.

Scenario 4 stack customization seem to be outside the philosophy of this library.
Scenario 5 seems to fit well but it remains to be seen if the exactMatch property would be useful

Therefor as a starting point, at the very least the API will look like this

```
Router(
  routes: [
    Route(
      path: '/sign-in', 
      builder: (context, params) => SignInPage(),
    ),
    Route(
      path: '/sign-in/phone', 
      builder: (context, params) => SignInWithPhonePage(),
    ),
  ]
)
```

## API design, adding home route

The two questions for the home route were:
  - Should the home route be part of the stack when navigating to '/products'
  - What happens when the user types the home route in our scenario since we didn't define that route

The two question are intertwined depending on how we approach it:
  1. Either the user is redirected to another page, eg: /dashboard in which case 
    the home route is only part of the stack when we are on /dashboard and is the stack ['/dashboard']
  2. Either '/' is not redirected, always defined and always part of the stack in which case
    we have to define what happens when we go to /sign-in
  3. Either '/' must be defined and only sometimes part of the stack, so when we are on '/dashboard' the stack can be either ['/', '/dashboard'] or just ['dashboard'] at user convenience
  4. Either '/' must be defined and never part of the stack exept when the route is home

2. can be eliminated because too complex with exceptions
4. doesn't allow much freedom
1. seems as a decent alternative in most use cases
3. offers a bit more freedom.

### Making it fit with the current API

1. can be implemented like this:

```
Router(
  Route(path: '/', redirectTo: '/dashboard'),
)
```
3. can be implemented like this

```
  Route(path: '/products', builder: ...),
  Route(path: '/', builder: ..., exactMatch: true / false ); // this is a required route
```

If exactMatch is true, it is part of the stack, if not then it's only part of the upstack when it's an exact match. This could apply to other routes as well.

*However what happens if the user doesn't specify '/' route ?*
 - the '/' route **must** be defined


Although this would give more freedom to the user this seems error prone. Having an home route that can potentially be on the stack of any route if the user doesn't pay attention, because it matches any child might bring more confusion than the alternative of just *always* redirecting the home route to another.

However having the option to redirectTo is a nice add-on to the API as it could help during development.
Each time we go to the route '/products' we are redirected to  '/dashboard' because the ProductPage widget doesn't exist yet


# Final API design: 

```
Router(
  routes: [
    Route(
      path: '/'
      exactMatch: true
    )
    Route(
      path: '/sign-in', 
      builder: (context, params) => SignInPage(),
      redirectTo: '/sign-in/phone'
    ),
    Route(
      path: '/sign-in/phone', 
      builder: (context, params) => SignInWithPhonePage(),
    ),
    ...
  ]
)

```