# Guards

## Goals

Guards should be reactive to reroute the user when the app state changes. Thankfully the Navigator 2.0 makes that easy.

*requirements:*
  - The guards must be reactive, reroute when the state changes
  - The guards must be self contained


## Scenario

For designing the API this fairly common and complex scenario will be taken as the base case to solve:

```
  '/loading' => user access this page when the authentication state is unknown
  '/sign-in' => user access this page when unauthenticated
  '/sign-in/phone' => user can access this nested page when unauthenticated 
  '/create-profile' => when user is authenticated but doesn't have a profile 
  '/dashboard' => when the user is authenticated & has a profile, the dashboard
  '/products' => when the user is authenticated & has a profile and clicked on a menu item to see the products
  '/products/:id' => when the user wants to see a specific product
```



## Option design 1. Reactive resolvers for all the routes:


```
  // Router(resolvers: [AuthResolver(), HasProfileResolver()], routes:...)

  class AuthResolver extends RouteResolver {
    // setting auth state will notify listeners
    state = 'unknown'; // can be 'authenticated', 'unauthenticated' or 'unknown'

    AuthResolver() {
      AuthRepository.authStateStream.listen((authState) => state = authState);
    }

    resolveOnStateChange(targetRoute) {

      switch(state) {
        case 'unknown':
          return '/loading';
        case 'authenticated':
          if (targetRoute.match('/sign-in')) return '/dashboard'; // *
          if (targetRoute.match('/loading')) return '/dashboard';
          else return targetRoute;
        case 'unauthenticated':
          if (targetRoute.match('/sign-in')) return targetRoute; // *
          else return '/sign-in;

          // * some checks are noted here for sake of being exhaustive but we don't have to make those
          // since if were previously not unauthenticated we wouldn't be on the sign-in route

          // the condition should look like this:

        case 'unknown':
          return '/loading';
        case 'authenticated':
          if (targetRoute.match('/loading')) return '/dashboard';
          else return targetRoute;
        case 'unauthenticated':
          return '/sign-in;

      }
    }
  }

  class HasProfileResolver extends RouteResolver {
    // setting auth state will notify listeners
    state = 'unknown'; // can be 'has', 'has_not' or 'unknown'

    HasProfileResolver() {
      ProfileRepository.profileState$.listen((profileState) => state = profileState);
    }

    resolve(targetRoute) {
      switch(state) {
        case 'unknown':
          return '/loading';
        case 'has':
          if (targetRoute.match('/loading')) return '/dashboard';
          if (targetRoute.match('/create-profile')) return '/dashboard';
          else return targetRoute;
        case 'has_not':
          return '/create-profile;
      }
    }
  }
```

Note: 
What happens when the client visit /sign-in and is authenticated but doesn't have a profile ?
  - AuthResolver resolves to /dashboard then HasProfileGuard resolves to /create-profile
What happens when Auth state is unknown at first 
  - AuthResolver resolves to /loading/sign-in, hasProfileResolver doesn't change that, then when the state changes we go on the needed route.
What happens if the user visits loading while authenticated
  - He will be redirected to the correct page depending if he has a profile

*advantages*:
  - easy to implement
*drawbacks*:
  - Possibly requires different loading routes as if we didn't have them and auth state would be unknown while hasProfile would be known, we would be redirected on the wrong page. It wouldn't make sens in this scenario though.
  - very manual we have to setup everything 
  - possibilities of loops
  - can't get on loading page if we are authenticated, but should we ?


## Option design 2. Future guards for specific routes:


```

Router(
  routes: [
    Route(path: '/sign-in', builder:..., resolvers: [unauthenticatedGuard]  ),
    Route(path: '/sign-in/phone', builder:...,  ), // we don't specify resolve here since it's child
    Route(path: '/dashboard', builder:..., resolvers: [authenticatedGuard, HasProfileGuard]  ),
    Route(path: '/loading', builder:... ),
    // ...
  ]
)

class AuthenticatedGuard extends RouteGuard {

  Future<bool> canActivate(targetRoute) {

    if (AuthRepository.currentState == 'unknown') {
      redirect('loading');
    }

    return AuthRepository.authStateStream
      .singleWhere((authState) => authState != 'unknown')
      .then((state) => state == 'authenticated' ? true : false )
      .then((can) {
        if (!can) {
          redirect('/sign-in')
        }
        return can;
      });
  }
}

```

*advantages*:
  - easier to see which route is affected
  - API similar to angular
*drawbacks*:
  - possibly hard to implement in the lib as we are redirecting inside the guard to a route which could have a future guard too. We can say that redirect will bypass resolvers though.
  - It's not clear if the API is easier for the end user compared to option 1.
  - in this scenario requires 2 different instance for authenticatedResolver and unauthenticatedResolver
  - lots of repetition for descendant of '/' that are not sign-in


## Option design 3 Reactive resolvers for specific routes

```
Router(
  routes: [
    Route(path: '/sign-in', builder:..., resolvers: [unauthenticatedGuard]  ),
    Route(path: '/sign-in/phone', builder:...,  ), // we don't specify resolve here since it's child
    Route(path: '/dashboard', builder:..., resolvers: [authenticatedGuard]  ),
    Route(path: '/loading', builder:... ),
    // ...
  ]
)
```
Solves some problems of option 2.

*advantages*:
  - no repetition for guards
  - can protect routes on specific paths from home without relying on the path
*drawbacks*
  - a more convoluted API
  - the guard API is the same as Option 2 which had it's own set of drawbacks


## Option 4 differentiate between guards and resolvers

A guard can output true and false on whether we should keep navigating


```
class AuthGuard {

  redirectOnStateChange(authState, currentRoute) {
    switch (authState) {
      case 'unknown': return '/loading';
      case 'authenticated': return '/dashboard';
      case 'unauthenticated': return 'sign-in';
    }
  }

  Future<bool> canActivate(targetRoute) {
    return AuthRepository.authStateStream
      .singleWhere((authState) => authState != 'unknown')
      .then((state) => state == 'authenticated' ? true : false );
  }
}
```

However this is scenario 1. but more complicated


## Scenarios conclusions

Scenario 3 and 1 are similar in API readability, although 3 might be a bit more standard on web frameworks.

For option 1. implementation the resolvers will output a route and that will become the current route. That's fairly simple to understand.

For option 3 (and 2) the the guards on the route and the parents will be checked in order, each returning a future. However since a redirect can be used inside the resolvers, we will redirect before the future completes and then redirect again. This seems more complicated.

Therefor option 1. is chosen


# Final API design


```
Router(
  home: '/dashboard',
  resolvers: [AuthResolver()],
  routes: ...
)
```